/******************************************************************************
 *  Copyright (C) 2002 John David Ratliff
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

// the size of the shared buffer
#define BUFFER_MAX	20

// the shared buffer
int pc_buffer[BUFFER_MAX];
int length;
int pos;

// mutexes are only used because pthread_cond_wait
// requires them. They do not lock each other out because
// only one uses the mutex.
pthread_mutex_t producer_m, consumer_m;

// condition variables for buffer
pthread_cond_t buffer_full, buffer_empty;

// the producer function
void *producer(void *value) {
	int nextp;
	time_t last = time(NULL);
	time_t now;

	for (;;) {
		// get a number between 0-99
		nextp = rand() % 100;

		// lock phony mutex for conditional
		pthread_mutex_lock(&producer_m);

		// add produced item to the buffer
		pc_buffer[pos++] = nextp;

		// increment the length of the array
		length++;

		// signal the buffer no longer empty
		pthread_cond_signal(&buffer_empty);

		// sleep on buffer full
		while (length == BUFFER_MAX) {
			pthread_cond_wait(&buffer_full, &producer_m);
		}

		// unlock phony mutex
		pthread_mutex_unlock(&producer_m);

		// only print once a minute
		// however, it rarely prints even once
		// the race condition often occurs WAY before 60 seconds
		now = time(NULL);
		if ((now - last) >= 60) {
			printf("producer: still working...\n");
			last = now;
		}
	}

	// don't really have anything to return...
	return NULL;
}

// the consumer function
void *consumer(void *value) {
	int nextc;

	// forever...
	for (;;) {
		// lock the consumer's phony mutex
		pthread_mutex_lock(&consumer_m);

		// sleep on buffer empty
		while (length == 0) {
			pthread_cond_wait(&buffer_empty, &consumer_m);
		}

		// grab the next item for consumption
		nextc = pc_buffer[pos--];
		length--;

		// signal buffer no longer full
		pthread_cond_signal(&buffer_full);

		// unlock this phony mutex
		pthread_mutex_unlock(&consumer_m);
	}

	// nothing really to return...
	return NULL;
}

int main() {
	pthread_t producer_t, consumer_t;

	// seed the random number generator
	srand(time(NULL));

	// erase the shared buffer
	memset(pc_buffer, 0, BUFFER_MAX * sizeof(int));
	length = 0;
	pos = 0;

	// initialize the mutexes/conditions
	pthread_mutex_init(&producer_m,NULL);
	pthread_mutex_init(&consumer_m,NULL);
	pthread_cond_init(&buffer_full,NULL);
	pthread_cond_init(&buffer_empty,NULL);

	// create the producer and consumer threads
	pthread_create(&producer_t,NULL,producer,NULL);
	pthread_create(&consumer_t,NULL,consumer,NULL);

	// exit the main thread, keep running producer/consumer
	pthread_exit(NULL);
}

